/*! \file anamod.h \ingroup functions
  \brief Prototypes for general module functions

  This file defines the functions for defining and querying 
  analysis modules.
*/
#ifndef MODULE_FUNCTIONS_H
#define MODULE_FUNCTIONS_H

#define ANAMOD_FORMAT_VERSION "1.0"

#include "petscmat.h"
#include "anamodtypes.h"

#define TRUTH(x) ( (x) ? PETSC_TRUE : PETSC_FALSE )
#define HASTOEXIST(h) \
  if (!h) {					 \
    printf("ERROR asking for unknown module\n"); \
    PetscFunctionReturn(1);			 \
  }
#define ANAMODCHECKVALID(i,c,s) {if (!i) SETERRQ(1,"Null pointer"); if (i->cookie!=c) SETERRQ1(1,"Not a valid <%s>",s);}

/*
 * Init/Final
 */
extern PetscErrorCode AnaModInitialize();
extern PetscErrorCode AnaModFinalize();
extern PetscErrorCode AnaModFinalize();
extern PetscErrorCode AnaModGetTypeName(int id,char **name);
extern PetscErrorCode AnaModGetTypeMySQLName(int id,char **name);

/*
 * Categories and modules
 */
extern PetscErrorCode AnaModRegisterStandardModules();
extern PetscErrorCode RegisterModule
(const char*,const char*,AnalysisDataType,
 PetscErrorCode(*f)(AnaModNumericalProblem,AnalysisItem*,int*,PetscTruth*));
extern PetscErrorCode DeRegisterCategory(char *cat);
extern PetscErrorCode DeregisterModules(void);
extern PetscErrorCode GetCategories(char***,int*);
extern PetscErrorCode CategoryGetModules(char*,char***,AnalysisDataType**,int**,int*);
extern PetscErrorCode HasComputeCategory(char *cat,PetscTruth *f);
extern PetscErrorCode HasComputeModule(char *cat,char *cmp,PetscTruth *f);

/*
 * Quantities
 */
extern PetscErrorCode GetDataID(const char*,const char*,int*,PetscTruth*);
extern PetscErrorCode GetDataType(const char*,const char*,AnalysisDataType*);
extern PetscErrorCode ComputeQuantity
(AnaModNumericalProblem,const char*,const char*,AnalysisItem*,int*,PetscTruth*);
extern PetscErrorCode ComputeQuantityByID
(AnaModNumericalProblem,int,int,AnalysisItem*,int*,PetscTruth*);
extern PetscErrorCode HasQuantity
(AnaModNumericalProblem,char*,char*,PetscTruth*);
extern PetscErrorCode HasQuantityByID
(AnaModNumericalProblem,int,AnalysisDataType,PetscTruth*);
extern PetscErrorCode RetrieveQuantity
(AnaModNumericalProblem,char*,char*,AnalysisItem*,PetscTruth*);
extern PetscErrorCode RetrieveQuantityByID
(AnaModNumericalProblem,int,int,AnalysisItem*,PetscTruth*);
extern PetscErrorCode QuantityAsString(AnalysisItem*,AnalysisDataType,char**);

/*
 * Execution control
 */
extern PetscErrorCode AnaModOptionsHandling(void);
extern PetscErrorCode AnaModShowOptions(MPI_Comm);
extern PetscErrorCode DeclareCategoryOptionFunction
(char *cat,PetscErrorCode(*f)(char*));
extern PetscErrorCode GetCategoryOptionFunction
(char *cat,PetscErrorCode(**f)(char*));
extern PetscErrorCode AnaModHasForcedSequentialComputation(PetscTruth*);
extern PetscErrorCode AnaModHasForcedExpensiveComputation(PetscTruth*);
extern PetscErrorCode AnaModGetSequentialMatrix
(Mat A,Mat *Awork,PetscTruth *mem,PetscTruth *local,PetscTruth *global);

/*
 * Tracing and output
 */
extern PetscErrorCode AnaModDeclareTraceFunction(PetscErrorCode(*)(void*,char*,va_list));
extern PetscErrorCode AnaModDeclareTraceContext(void*);
extern PetscErrorCode AnaModTraceMessage(const char *fmt,...);
extern PetscErrorCode AnaModHasTrace(PetscTruth *flg);
extern PetscErrorCode AnaModSetTraceArrays(PetscTruth f);
extern PetscErrorCode AnaModTraceArrays(PetscTruth *f);
#if defined(HAVE_NMD)
extern PetscErrorCode TabReportModules(char**,int);
extern PetscErrorCode TabReportValues(NMD_metadata,char**,char**,int);
#else
extern PetscErrorCode TabReportModules(Mat,char**,char**,int);
#endif
extern PetscErrorCode PurgeAttachedArrays(Mat A);

/*
 * Feature sets
 */
typedef struct FeatureSet_* FeatureSet;
typedef struct FeatureValues_* FeatureValues;
extern PetscErrorCode NewFeatureSet(FeatureSet*);
extern PetscErrorCode DeleteFeatureSet(FeatureSet);
extern PetscErrorCode AddToFeatureSet(FeatureSet,char*,char*,int*);
extern PetscErrorCode NewFeatureValues(FeatureValues*);
extern PetscErrorCode DeleteFeatureValues(FeatureValues);
extern PetscErrorCode InstantiateFeatureSet
(AnaModNumericalProblem,FeatureSet,FeatureValues);
extern PetscErrorCode GetFeatureValue
(FeatureValues,int,AnalysisItem*,PetscTruth*);
extern PetscErrorCode AnaModSetRetrievalFunction
(PetscErrorCode(*)(void*,char*,char*,AnalysisItem*,AnalysisDataType*,PetscTruth*));
extern PetscErrorCode AnaModGetRetrievalFunction
(PetscErrorCode(**)(void*,char*,char*,AnalysisItem*,AnalysisDataType*,PetscTruth*),PetscTruth*);

/*
 * Utils
 */
extern PetscErrorCode CreateIntArray(const char*,int,IntArray*);
extern PetscErrorCode DeleteIntArray(IntArray);
extern PetscErrorCode IntArrayAdd(IntArray,int,int*);
extern PetscErrorCode IntArraySetAt(IntArray,int,int);
extern PetscErrorCode IntArrayTryGetAt(IntArray,int,int*,PetscTruth*);
extern PetscErrorCode IntArrayGetAt(IntArray,int,int*);
extern PetscErrorCode IntArrayGetFill(IntArray,int*);

extern PetscErrorCode CreateStringArray(const char*,int,StringArray*);
extern PetscErrorCode DeleteStringArray(StringArray);
extern PetscErrorCode StringArrayAdd(StringArray,char*,int*);
extern PetscErrorCode StringArraySetAt(StringArray,int,char*);
extern PetscErrorCode StringArrayTryGetAt(StringArray,int,char**,PetscTruth*);
extern PetscErrorCode StringArrayGetAt(StringArray,int,char**);
extern PetscErrorCode StringArrayGetFill(StringArray,int*);

extern PetscErrorCode CreateAnalysisItemArray(char*,int,AnalysisItemArray*);
extern PetscErrorCode DeleteAnalysisItemArray(AnalysisItemArray);
extern PetscErrorCode AnalysisItemArrayAdd(AnalysisItemArray,AnalysisItem,int*);
extern PetscErrorCode AnalysisItemArraySetAt(AnalysisItemArray,int,AnalysisItem);
extern PetscErrorCode AnalysisItemArrayTryGetAt(AnalysisItemArray,int,AnalysisItem*,PetscTruth*);
extern PetscErrorCode AnalysisItemArrayGetAt(AnalysisItemArray,int,AnalysisItem*);

extern PetscErrorCode CreateAnalysisDataTypeArray(char*,int,AnalysisDataTypeArray*);
extern PetscErrorCode DeleteAnalysisDataTypeArray(AnalysisDataTypeArray);
extern PetscErrorCode AnalysisDataTypeArrayAdd(AnalysisDataTypeArray,AnalysisDataType,int*);
extern PetscErrorCode AnalysisDataTypeArraySetAt(AnalysisDataTypeArray,int,AnalysisDataType);
extern PetscErrorCode AnalysisDataTypeArrayTryGetAt(AnalysisDataTypeArray,int,AnalysisDataType*,PetscTruth*);
extern PetscErrorCode AnalysisDataTypeArrayGetAt(AnalysisDataTypeArray,int,AnalysisDataType*);

/*
 * Logging
 */
extern PetscErrorCode CategoryLogEventRegister(char *cat,int icat);

#endif
