#ifndef NMD_H
#define NMD_H 1

/*! \page types NMD Data Types

  Most NMD datatypes have two different definitions, depending on 
  whether Petsc is used or not.

  NMDTruth is only used for success parameters in routines
  such as NMDGetValue(). It has possible values NMDTrue and NMDFalse.
*/

#if defined(NMD_HAVE_PETSC)
#include "petsc.h"
typedef PetscReal NMDRealtype;
typedef PetscTruth NMDTruth;
#define NMDTrue PETSC_TRUE
#define NMDFalse PETSC_FALSE
typedef PetscErrorCode NMDErrorCode;
#else
typedef double NMDRealtype;
typedef int NMDTruth;
#define NMDTrue 1
#define NMDFalse 0
typedef int NMDErrorCode;
#endif

typedef enum {
  NMDInvalid, NMDInt, NMDReal, NMDString, NMDIntarray, NMDRealarray
}  NMDDataType;
extern const char *typenames[];

typedef struct NMD_metadata_item_* NMD_metadata_item;
typedef struct NMD_metadata_category_* NMD_metadata_category;
typedef struct NMD_metadata_* NMD_metadata;
typedef struct NMD_object_* NMD_object;

#define NMDCOOKIE 32897432
#define CHECKHASNMDCOOKIE(obj) { if (!obj) NMD_ERR_REPORT("Null object"); if (((NMD_object)(obj))->cookie!=NMDCOOKIE) NMD_ERR_REPORTi("Object has invalid cookie",((NMD_object)(obj))->cookie); }

/*
 * Object
 */
extern NMDErrorCode NMDCreateObject(NMD_metadata*);
extern NMDErrorCode NMDDestroyObject(NMD_metadata);
extern NMDErrorCode NMDViewObject(NMD_metadata);
extern NMDErrorCode NMDBuildObjectStructure(NMD_metadata);
extern NMDErrorCode NMDDestroyObjectStructure(NMD_metadata);
extern NMDErrorCode NMDCloneObjectStructure(NMD_metadata,NMD_metadata*);
extern NMDErrorCode NMDCloneObject(NMD_metadata,NMD_metadata);
extern NMDErrorCode NMDReportObject
(NMD_metadata,NMDTruth,char**,char**,char,char,char);
/*
 * Category
 */
extern NMDErrorCode NMDObjectAllocateNewCategory
(NMD_metadata,const char*,NMD_metadata_category*);
extern NMDErrorCode NMDObjectTryGetCategory
(NMD_metadata,const char*,NMD_metadata_category*,NMDTruth*);
extern NMDErrorCode NMDObjectGetCategory(NMD_metadata,char*,NMD_metadata_category*);
extern NMDErrorCode NMDObjectGetOrCreateCategory
(NMD_metadata obj,char *cat,NMD_metadata_category *ctg);
extern NMDErrorCode NMDRemoveCategory(NMD_metadata,const char*);
extern NMDErrorCode NMDCopyCategory(NMD_metadata_category,NMD_metadata_category);
extern NMDErrorCode NMDGetCategories(NMD_metadata,int*,char***);
/*
 * Component
 */
extern NMDErrorCode NMDCategoryAllocateNewComponent
(NMD_metadata_category,const char*,NMDDataType,NMD_metadata_item*);
extern NMDErrorCode NMDComponentDestroy(NMD_metadata_item);
extern NMDErrorCode NMDCategoryCreateComponent(NMD_metadata,char*,char*);
extern NMDErrorCode NMDCategoryGetComponents
(NMD_metadata,char*,int*,char***,NMDDataType**);
extern NMDErrorCode NMDCategoryGetOrCreateComponent
(NMD_metadata_category,char*,NMDDataType,NMD_metadata_item*);
extern NMDErrorCode NMDCategoryTryGetComponent
(NMD_metadata_category,const char*,NMD_metadata_item*,NMDTruth*);
extern NMDErrorCode NMDObjectHasCategoryComponent
(NMD_metadata,char*,char*,NMDTruth*);
extern NMDErrorCode NMDObjectEnsureCategoryComponent
(NMD_metadata,char*,char*,NMDDataType,NMDTruth*);
extern NMDErrorCode NMDCategoryGetComponent
(NMD_metadata_category,char*,NMD_metadata_item*);
extern NMDErrorCode NMDGetCategoryIGetComponents(NMD_metadata,int,int*,char***,NMDDataType**);
/* values */
extern NMDErrorCode NMDSetValue(NMD_metadata,const char*,const char*,NMDDataType,void*);
extern NMDErrorCode NMDComponentSetValue(NMD_metadata_item,NMDDataType,void*);
extern NMDErrorCode NMDComponentUnsetValue(NMD_metadata_item);
extern NMDErrorCode NMDSetArrayValue
(NMD_metadata,const char*,const char*,NMDDataType,void*,int);
extern NMDErrorCode NMDComponentSetArrayValue
(NMD_metadata_item,NMDDataType,void*,int);
extern NMDErrorCode NMDCopyArrayValue
(NMD_metadata,const char*,const char*,NMDDataType,void*,int);
extern NMDErrorCode NMDGetValue
(NMD_metadata,const char*,const char*,NMDDataType*,void*,NMDTruth*);
extern NMDErrorCode NMDGetArrayValue
(NMD_metadata,char*,char*,NMDDataType*,void*,int*,NMDTruth*);
extern NMDErrorCode NMDCopyItemValues(NMD_metadata_item,NMD_metadata_item);
extern NMDErrorCode NMDGetDataType(NMD_metadata,char*,char*,NMDDataType*t);
extern NMDErrorCode NMDIsArrayType(NMDDataType type,NMDTruth*);
extern NMDErrorCode NMDUnsetValue(NMD_metadata,const char*,const char*);

extern PetscErrorCode NMDGetTypeMySQLName(NMDDataType,char**);

#if defined(NMD_HAVE_HDF5)
/*
 * hdf5 dump
 */
extern NMDErrorCode NMDObjectDumpToHDF5(NMD_metadata obj,char *fname);
#endif

/*
 * our own string library
 */
typedef struct NMD_string_* NMD_string;

extern NMDErrorCode NMDStringCreate(char*,NMD_string*);
extern NMDErrorCode NMDStringDestroy(NMD_string);
extern NMDErrorCode NMDStringGetString(NMD_string str,char **t);
extern NMDErrorCode NMDStringConcat
(char,NMD_string,char,NMD_string,char,NMD_string*);

#define NMD_ERR_RETURN(a) if (a) {printf("Error in <%s>\n  line %d of file %s.\n",__FUNCT__,__LINE__,__FILE__); return a;}
#define NMD_ERR_REPORT(a)     {printf("Error in <%s>\n  line %d of file %s:\n\n||  %s  ||  \n\n",            __FUNCT__,__LINE__,__FILE__,a); return 1;}
#define NMD_ERR_REPORTs(a,b)     {printf("Error in <%s>\n  line %d of file %s:\n\n||  %s: %s  ||\n\n",     __FUNCT__,__LINE__,__FILE__,a,b); return 1;}
#define NMD_ERR_REPORTss(a,b,c)						    {printf("Error in <%s>\n  line %d of file %s:\n\n||  %s: %s,%s  ||\n\n", 	    __FUNCT__,__LINE__,__FILE__,a,b,c);	return 1;}
#define NMD_ERR_REPORTi(a,b)     {printf("Error in <%s>\n  line %d of file %s:\n\n||  %s: %d  ||\n\n", 	    __FUNCT__,__LINE__,__FILE__,a,b); return 1;}
#define NMD_ERR_REPORTii(a,b,c)     {printf("Error in <%s>\n  line %d of file %s:\n\n||  %s: %d,%d  ||\n\n", 	    __FUNCT__,__LINE__,__FILE__,a,b,c); return 1;}
#if defined(NMD_HAVE_PETSC)
#define NMD_MALLOC(a,b,c,d)	 \
  { NMDErrorCode merr; merr = PetscMalloc((b)*sizeof(c),&a); CHKERRQ(merr); \
    merr = PetscMemzero(a,(b)*sizeof(c)); CHKERRQ(merr); CHKMEMQ; }
#define NMD_FREE(a) {NMDErrorCode merr; merr = PetscFree(a); CHKERRQ(merr); }
#define NMD_STRDUP(a,b) { NMDErrorCode merr; merr = PetscStrallocpy(a,&b); CHKERRQ(merr); }
#else
#define NMD_MALLOC(a,b,c,d)	 \
  { a = (c*)malloc((b)*sizeof(c));		   \
    if (!a) NMD_ERR_REPORTs("Could not allocate",d); \
    memset(a,0,(b)*sizeof(c)); }
#define NMD_FREE(a) {free(a);}
#define NMD_STRDUP(a,b) b = strdup(a);
#endif

#endif

